const path = require ('path')
const _ = require ('lodash')

const { getOptions } = require ('loader-utils')

const postcss = require ('postcss')
const postcssrc = require ('postcss-load-config')

class SyntaxError extends Error {
  constructor(error) {
    super (error)

    const { line, column, reason } = error

    this.name = 'SyntaxError'

    this.message = `${this.name}\n\n`

    if (typeof line !== 'undefined') {
      this.message += `(${line}:${column}) `
    }

    this.message += `${reason}`

    const code = error.showSourceCode ()

    if (code) {
      this.message += `\n\n${code}\n`
    }

    this.stack = false
  }
}

class Warning extends Error {
  constructor(warning) {
    super (warning)

    const { text, line, column } = warning

    this.name = 'Warning'

    this.message = `${this.name}\n\n`

    if (typeof line !== 'undefined') {
      this.message += `(${line}:${column}) `
    }

    this.message += `${text}`

    this.stack = false
  }
}


function loader(css, map, meta) {
  const options = Object.assign ({}, getOptions (this))

  const cb = this.async ()
  const file = this.resourcePath

  const sourceMap = options.sourceMap

  Promise.resolve ().then (() => {
    const rc = {
      path: path.dirname (file),
      ctx: {
        file: file,
        options: {}
      }
    }

    options.config = options.config || {}

    if (options.config.ctx) {
      rc.ctx.options = _.isFunction (options.config.ctx) ? options.config.ctx.call (this) : options.config.ctx
    }

    rc.ctx.webpack = this
    let current = {}
    let root = {}

    try {
      current = postcssrc.sync (rc.ctx, rc.path)
    } catch (e) {

    }
    try {
      if (options.config.path) {
        root = postcssrc.sync (rc.ctx, path.resolve (options.config.path))
      }
    } catch (e) {

    }

    return Object.assign ({},
      rc.ctx,
      root,
      current,
      {
        plugins: [
          ...(rc.ctx.options.plugins || []),
          ...(root.plugins || []),
          ...(current.plugins || [])
        ]
      }
    )
  }).then ((config) => {
    if (!config) {
      config = {}
    }

    if (config.file) {
      this.addDependency (config.file)
    }

    if (config.options.to) {
      delete config.options.to
    }
    if (config.options.from) {
      delete config.options.from
    }

    let plugins = config.plugins || []

    let options = Object.assign ({
      from: file,
      map: sourceMap
        ? sourceMap === 'inline'
          ? { inline: true, annotation: false }
          : { inline: false, annotation: false }
        : false
    }, config.options)

    if (options.parser === 'postcss-js') {
      css = this.exec (css, this.resource)
    }

    if (typeof options.parser === 'string') {
      options.parser = require (options.parser)
    }

    if (typeof options.syntax === 'string') {
      options.syntax = require (options.syntax)
    }

    if (typeof options.stringifier === 'string') {
      options.stringifier = require (options.stringifier)
    }

    if (config.exec) {
      css = this.exec (css, this.resource)
    }

    if (sourceMap && typeof map === 'string') {
      map = JSON.parse (map)
    }

    if (sourceMap && map) {
      options.map.prev = map
    }

    return postcss (plugins)
      .process (css, options)
      .then ((result) => {
        let { css, map, root, processor, messages } = result

        result.warnings ().forEach ((warning) => {
          this.emitWarning (new Warning (warning))
        })

        messages.forEach ((msg) => {
          if (msg.type === 'dependency') {
            this.addDependency (msg.file)
          }
        })

        map = map ? map.toJSON () : null

        if (map) {
          map.file = path.resolve (map.file)
          map.sources = map.sources.map ((src) => path.resolve (src))
        }

        if (!meta) {
          meta = {}
        }

        const ast = {
          type: 'postcss',
          version: processor.version,
          root
        }

        meta.ast = ast
        meta.messages = messages

        if (this.loaderIndex === 0) {
          cb (null, `module.exports = ${JSON.stringify (css)}`, map)
          return null
        }

        cb (null, css, map, meta)

        return null
      })
  }).catch ((err) => {
    if (err.file) {
      this.addDependency (err.file)
    }

    return err.name === 'CssSyntaxError'
      ? cb (new SyntaxError (err))
      : cb (err)
  })
}

module.exports = loader
